function [err, bit_trim_info]=unpack_ifgs(bit_trim_file_in, bit_trim_file_out)
% Copyright (c) 2007-2014 University of Wisconsin Regents.
%
% This program is free software: you can redistribute it and/or modify
%    it under the terms of the GNU General Public License as published by
%    the Free Software Foundation, either version 3 of the License, or
%    (at your option) any later version.
%
%    This program is distributed in the hope that it will be useful,
%    but WITHOUT ANY WARRANTY; without even the implied warranty of
%    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
%    GNU General Public License for more details.
%
%    You should have received a copy of the GNU General Public License
%    along with this program.  If not, see <http://www.gnu.org/licenses/>.
%
%
% Unpack interferograms by applying bit-trim parameters.
% This routine makes assumptions about the names of fields in packets, and could 
% break if field names are changed in a newer version of the XML database.
    
set_globals

err = 0;
bit_trim_info.source = 'none';
bands = {'lw', 'mw', 'sw'};
scenes = {'es', 'ds', 'ict'};
diag_names = {'lw_diag', 'mw_diag', 'sw_diag'};

% if there are any diagnostic mode ifgs, unpack them
num_diag_packets = 0;
for diag_name = diag_names
    diag_name = char(diag_name);
    
    if(g_metadata.packet_stats.num_packets.(diag_name) > 0)
    
        num_diag_packets = num_diag_packets + g_metadata.packet_stats.num_packets.(diag_name);
        
        g_ifgs.(diag_name).for = int8(g_packets.(diag_name).ScnInfo_FOR);
        g_ifgs.(diag_name).sweep_dir = int8(g_packets.(diag_name).ScnInfo_Sweep_Dir);
        g_ifgs.(diag_name).dnum = g_packets.(diag_name).header_time;
        g_ifgs.(diag_name).chan = arrayfun(@decode_diag_mode_channel_select, g_packets.(diag_name).FilterStatLowerReg_Diag_Mode_Chan_Sel);
        g_ifgs.(diag_name).real = g_packets.(diag_name).DIA_DetectorData;
       
        g_metadata.ifgs.(diag_name).for = g_metadata.packets.(diag_name).ScnInfo_FOR;
        g_metadata.ifgs.(diag_name).sweep_dir = g_metadata.packets.(diag_name).ScnInfo_Sweep_Dir;
        g_metadata.ifgs.(diag_name).dnum = g_metadata.packets.(diag_name).header_time;
        g_metadata.ifgs.(diag_name).chan = g_metadata.packets.(diag_name).FilterStatLowerReg_Diag_Mode_Chan_Sel;        
        g_metadata.ifgs.(diag_name).real = g_metadata.packets.(diag_name).DIA_DetectorData;
        g_metadata.ifgs.(diag_name).dim = g_metadata.packets.(diag_name).dim;     
    
    end
end
if(num_diag_packets > 0)
    fprintf('  Unpacked %d diagnostic interferograms\n', num_diag_packets)
end

% determine if there are any interferogram packets to unpack
num_ifg_packets = 0;
for band = {'lw', 'mw', 'sw'}
    for fov = 1:9
        for scene = {'ds', 'es', 'ict'}
            num_ifg_packets = num_ifg_packets + g_metadata.packet_stats.num_packets.(strcat('ifg_',char(band),'_fov', int2str(fov),'_',char(scene)));
        end
    end
end
if(num_ifg_packets == 0)
    if(num_diag_packets == 0)
        fprintf('  No interferograms to unpack\n')
    end
    return
end

% read bit-trim values from file if requested
if(~isempty(bit_trim_file_in))
    bit_trim_info.source = 'file';
    bit_trim_info.input_file = bit_trim_file_in;
    [bit_trim_info.BitTrimIndex, bit_trim_info.BitTrimBitsRetained] = read_bit_trim_values(bit_trim_file_in);
    fprintf('  Input bit trim H5 file: %s\n', bit_trim_file_in)

% else calculate bit-trim values from parameters in engineering packet
else
    
    % need at least one eng packet
    if(g_metadata.packet_stats.num_packets.four_min_eng < 1)
        fprintf('  Found no engineering packets. Cannot unpack interferogram RDR data without bit-trim values\n')
        err = -10; 
        return
    end
    
    bit_trim_info.source = 'data_stream';
    
    % FIXME: should only print warning if there are differences in bit-trim values from different packets
    % Note that there is no obvious way to correctly handle datasets that span changes in bit-trim
    % values because both engineering packets may or may not be present.
    if(g_metadata.packet_stats.num_packets.four_min_eng > 1)
        fprintf('  WARNING: Found %d engineering packets. Will use bit-trim values from the first engineering packet to unpack interferogram RDR data\n', g_metadata.packet_stats.num_packets.four_min_eng)
    end
    
    % calculate parameters required by unpacking routine from engineering values,
    % for all bands and bit-trim zones
    eng_packet_idx = 1;
    bit_trim_info.BitTrimIndex = zeros(16, 3, 'double');
    bit_trim_info.BitTrimBitsRetained = zeros(16, 3, 'double');
    for band_idx=1:3
         for zone_idx=1:16
            zone = int2str(zone_idx - 1);  % zones are from 0 to 15
            bit_trim_info.BitTrimIndex(zone_idx, band_idx) = ...
                g_packets.four_min_eng.('EndSampleIndex')(zone_idx, band_idx, eng_packet_idx);
            bit_trim_info.BitTrimBitsRetained(zone_idx, band_idx) = ...
                g_packets.four_min_eng.('FIRAccumulator_StopBit')(zone_idx, band_idx, eng_packet_idx) ...
                - g_packets.four_min_eng.('FIRAccumulator_StartBit')(zone_idx, band_idx, eng_packet_idx) + 1;
        end
    end
end

% write bit-trim values to file if requested
if(~isempty(bit_trim_file_out))
    write_bit_trim_values(bit_trim_file_out, bit_trim_info.BitTrimIndex, bit_trim_info.BitTrimBitsRetained);
    fprintf('  Output bit trim H5 file: %s\n', bit_trim_file_out)
end

% unpack interferograms
num_ifgs_unpacked = 0;
g_metadata.ifgs.dim.fov = 9;
for scene_idx = 1:3
    scene = scenes{scene_idx};
    
    for band_idx = 1:3
        band = bands{band_idx};
        band_and_scene = [band, '_', scene];
        
        % determine max number of packets for any FOV
        max_packets = 0;
        for fov = 1:9
            varname = ['ifg_',band,'_fov',int2str(fov),'_',scene];
            max_packets = max(max_packets, g_metadata.packet_stats.num_packets.(varname));
        end
        
        % allocate FOR, sweep_dir and dnum arrays to hold the max number of packets for any FOV.
        % This is more to get non-zero missing sentinels than for efficiency.
        % FUTURE: could pre-allocate real and imag arrays, but would need to know size of array
        % returned by bit_unpack_all.
        if(max_packets > 0)
            g_ifgs.(band_and_scene).for = ones(9, max_packets, 'int8') * intmin('int8');
            g_ifgs.(band_and_scene).sweep_dir = ones(9, max_packets, 'int8') * intmin('int8');
            g_ifgs.(band_and_scene).dnum = NaN(9, max_packets, 'double');
        end
                
        for fov = 1:9
            varname = ['ifg_',band,'_fov',int2str(fov),'_',scene];
            if(g_metadata.packet_stats.num_packets.(varname) == 0)
                continue
            end
            fieldname = ['DetectorData'];
            s = size(g_packets.(varname).(fieldname));

            for packet_idx = 1:g_metadata.packet_stats.num_packets.(varname)

                % assuming first half of points is real data, second half is imag
                g_ifgs.(band_and_scene).('real')(:, fov, packet_idx) = bit_unpack_all( double(g_packets.(varname).(fieldname)(1:s(1)/2, packet_idx)),bit_trim_info.BitTrimIndex(:,band_idx),bit_trim_info.BitTrimBitsRetained(:,band_idx));
                g_ifgs.(band_and_scene).('imag')(:, fov, packet_idx) = bit_unpack_all( double(g_packets.(varname).(fieldname)(s(1)/2+1:s(1), packet_idx)),bit_trim_info.BitTrimIndex(:,band_idx),bit_trim_info.BitTrimBitsRetained(:,band_idx));                
                num_ifgs_unpacked = num_ifgs_unpacked + 1;
                                                        
            end
                
            % set FOR, sweep dir and dnum (time)
            packet_str = ['ifg_', band, '_fov', int2str(fov), '_', scene];
            num_packets = g_metadata.packets.(packet_str).dim.([packet_str '_packet']);
            if(num_packets > 0)
                g_ifgs.(band_and_scene).for(fov, 1:num_packets) = int8(g_packets.(packet_str).ScanInfo_Field_of_Regard(:));
                g_ifgs.(band_and_scene).sweep_dir(fov, 1:num_packets) = int8(g_packets.(packet_str).ScanInfo_Sweep_Dir(:));
                g_ifgs.(band_and_scene).dnum(fov, 1:num_packets) = g_packets.(packet_str).header_time(:);
            end
        end

        % set dims
        if(isfield(g_ifgs, band_and_scene) && isfield(g_ifgs.(band_and_scene), 'real'))
            [g_metadata.ifgs.dim.([band_and_scene, '_points']), x, g_metadata.ifgs.dim.([band_and_scene, '_ifg'])] = ...
                size(g_ifgs.(band_and_scene).('real'));
            g_metadata.ifgs.(band_and_scene).('real').dim = ...
                {[band_and_scene, '_points'], 'fov', [band_and_scene, '_ifg']};
            g_metadata.ifgs.(band_and_scene).('imag').dim = ...
                {[band_and_scene, '_points'], 'fov', [band_and_scene, '_ifg']};
            g_metadata.ifgs.(band_and_scene).('for').dim = ...
                {'fov', [band_and_scene, '_ifg']};
            g_metadata.ifgs.(band_and_scene).('sweep_dir').dim = ...
                {'fov', [band_and_scene, '_ifg']};
            g_metadata.ifgs.(band_and_scene).('dnum').dim = ...
                {'fov', [band_and_scene, '_ifg']};
        end
    end
end

fprintf('  Unpacked %d complex RDR interferograms\n', num_ifgs_unpacked)
        
end %unpack_ifgs



function write_bit_trim_values(bit_trim_file_out, BitTrimIndex, BitTrimBitsRetained)
    save(bit_trim_file_out, 'BitTrimIndex', 'BitTrimBitsRetained', '-mat', '-v7.3');
end

function [BitTrimIndex, BitTrimBitsRetained]=read_bit_trim_values(bit_trim_file_in)
    load(bit_trim_file_in, 'BitTrimIndex', 'BitTrimBitsRetained', '-mat');
end

% Decode "channel" from binary values in the "Diagnostic Mode Channel Select" field
% from a diagnostic packet. Following the documentation in the Command and Data
% Dictionary, 19 Apr 2013. The "channel" value is believed to be the same as FOV index.
function channel = decode_diag_mode_channel_select(chan_select)
    switch chan_select     
        %                    binary
      case 24              % 11000
        channel = int8(1);
      case 25              % 11001
        channel = int8(2);
      case 26              % 11010
        channel = int8(3);
      case 27              % 11011
        channel = int8(4);
      case 20              % 10100
        channel = int8(5);
      case 21              % 10101
        channel = int8(6);
      case 22              % 10110
        channel = int8(7);
      case 23              % 10111
        channel = int8(8);
      case 12              % 01100
        channel = int8(9);
      otherwise
        channel = intmin('int8');
    end
end
        
